#include <Arduino.h>
#include "control.h"
#include "speaker.h"
#include "rtc.h"

#define ENC1 11
#define ENC2 10
#define ENC_BTN 12
#define ALARM_BTN 33
#define ALARM_LED 32

#define OPTION1 39
#define OPTION2 38
#define OPTION3 41
#define OPTION4 40

#define IDLE_TIME 5000 // ms
#define LONG_BTN 1000 // ms
#define DEBOUNCE_DELAY 100 // ms

static unsigned char _mode;
static char _encoder_step;
static unsigned long _enc_btn_ms;
static unsigned long _alarm_btn_ms;
static unsigned long _idle_ms;

static button_state _alarm_btn_state;
static button_state _enc_btn_state;

static unsigned char _modified_value;

void control_init() {
    PCICR |= 0b00000001; // We activate the interrupts of the PB port
    PCMSK0 |= 0b00110000; // We activate the interrupts on pins D10, D11 
    pinMode(ENC1, INPUT_PULLUP);
    pinMode(ENC2, INPUT_PULLUP);
    pinMode(ENC_BTN, INPUT_PULLUP);
    pinMode(ALARM_BTN, INPUT_PULLUP);
    pinMode(OPTION1, INPUT_PULLUP);
    pinMode(OPTION2, INPUT_PULLUP);
    pinMode(OPTION3, INPUT_PULLUP);
    pinMode(OPTION4, INPUT_PULLUP);
    pinMode(ALARM_LED, OUTPUT);
    _mode = SHOW_TIME;
    _encoder_step = 0;
    _enc_btn_state = idle;
    _idle_ms = 0;
    _alarm_btn_ms = 0;
    _alarm_btn_state = idle;
    digitalWrite(ALARM_LED, rtc.getAlarmFlags());
}

void increment60(unsigned char& value) {
  value++;
  if ((value & 0x0F) % 10 == 0)
    value += 6;
  if (value == 0x60)
    value = 0x00;
}

void decrement60(unsigned char& value) {
  value--;
  if (value == 0xFF)
    value = 0x59;
  if ((value & 0x0F) % 16 == 15)
    value -= 6;
}

void increment12(unsigned char& value) {
  value++;
  if ((value & 0x0F) % 10 == 0)
    value += 6;
  if (value == 0x13)
    value = 0x01;
}

void decrement12(unsigned char& value) {
  value--;
  if (value == 0x00)
    value = 0x12;
  if ((value & 0x0F) % 16 == 15)
    value -= 6;
}

void increment24(unsigned char& value) {
  value++;
  if ((value & 0x0F) % 10 == 0)
    value += 6;
  if (value == 0x24)
    value = 0x00;
}

void decrement24(unsigned char& value) {
  value--;
  if (value == 0xFF)
    value = 0x23;
  if ((value & 0x0F) % 16 == 15)
    value -= 6;
}

inline void encoder_increment() {
    switch (_mode) {
        case SET_ALARM_HOUR:
            increment24(_modified_value);
            break;
        case SET_ALARM_MINUTE:
        case SET_ALARM_SECOND:
            increment60(_modified_value);
            break;
    }
}

inline void encoder_decrement() {
    switch (_mode) {
        case SET_ALARM_HOUR:
            decrement24(_modified_value);
            break;
        case SET_ALARM_MINUTE:
        case SET_ALARM_SECOND:
            decrement60(_modified_value);
            break;
    }
}

inline signed char read_encoder() {
    static char buff;

    buff <<= 1;
    if (digitalRead(ENC1) == LOW)
        buff += 1;
    buff <<= 1;
    if (digitalRead(ENC2) == LOW)
        buff += 1;

    switch (buff & 0x0F) {
        case 0b00000001: return 1;
        case 0b00000010: return -1;
        case 0b00000100: return -1;
        case 0b00000111: return 1;
        case 0b00001000: return 1;
        case 0b00001011: return -1;
        case 0b00001101: return -1;
        case 0b00001110: return 1;
    }
    return 0;
}

ISR (PCINT0_vect) {
    _encoder_step += read_encoder();
    if (_encoder_step > 3) {
        _encoder_step = 0;
        encoder_increment();
    }
    if (_encoder_step < -3) {
        _encoder_step = 0;
        encoder_decrement();
    }
}

inline button_state read_encoder_btn() {
    if (digitalRead(ENC_BTN) == LOW) {
        if (millis() - _enc_btn_ms > LONG_BTN) {
            return long_pressed;
        }
        if (millis() - _enc_btn_ms > DEBOUNCE_DELAY) {
            return pressed;
        } else {
            return idle;
        }
    } else {
        _enc_btn_ms = millis();
        return idle;
    }
}

inline button_state read_alarm_btn() {
    if (digitalRead(ALARM_BTN) == LOW) {
        if (millis() - _alarm_btn_ms > LONG_BTN) {
            return long_pressed;
        } else {
            return pressed;
        }
    } else {
        _alarm_btn_ms = millis();
        return idle;
    }
}

void select_mode(unsigned char mode) {
      _mode = mode;
}

void select_next_mode() {
      Serial.print('*');
}


void inline alarm_btn_press() {
    speaker_mute();
    _mode = SHOW_DATE;
    _idle_ms = millis() + IDLE_TIME;
}

void inline alarm_btn_long_press() {
    _mode = SHOW_ALARM;
    _idle_ms = millis() + IDLE_TIME;
    bool alarm_enabled = rtc.getAlarmFlags();
    if (alarm_enabled) {
        rtc.setAlarmFlags(0);
        digitalWrite(ALARM_LED, 0);
    } else {
        rtc.setAlarmFlags(1);
        digitalWrite(ALARM_LED, 1);
    }
}

void inline save_alarm_hour() {
    if (_modified_value != rtc.getAlarmHour()) {
        rtc.setAlarmHour(_modified_value);
    }
}

void inline save_alarm_minute() {
    if (_modified_value != rtc.getAlarmMinute()) {
        rtc.setAlarmMinute(_modified_value);
    }
}

void inline save_alarm_second() {
    if (_modified_value != rtc.getAlarmSecond()) {
        rtc.setAlarmSecond(_modified_value);
    }
}

void inline enc_btn_press() {
    switch (_mode) {
        case SHOW_TIME:
        case SHOW_DATE:
            _mode = SET_ALARM_HOUR;
            _modified_value = rtc.getAlarmHour();
            break;
        case SET_ALARM_HOUR:
            save_alarm_hour();
            _mode = SET_ALARM_MINUTE;
            _modified_value = rtc.getAlarmMinute();
            break;
        case SET_ALARM_MINUTE:
            save_alarm_minute();
            _mode = SET_ALARM_SECOND;
            _modified_value = rtc.getAlarmSecond();
            break;
        case SET_ALARM_SECOND:
            save_alarm_second();
            _mode = SHOW_TIME;
            break;
    }
}

unsigned char get_mode() {
    button_state alarm_btn = read_alarm_btn();
    button_state encoder_btn = read_encoder_btn();

    if (_enc_btn_state != encoder_btn) {
        _enc_btn_state = encoder_btn;
        switch (encoder_btn) {
            case pressed:
                enc_btn_press();
                break;
        }
    }

    if (_mode == SET_ALARM_HOUR || _mode == SET_ALARM_MINUTE || _mode == SET_ALARM_SECOND) {
        return _mode;
    }
    
    if (_alarm_btn_state != alarm_btn) {
        _alarm_btn_state = alarm_btn;
        switch (alarm_btn) {
            case pressed:
                alarm_btn_press();
                break;
            case long_pressed:
                alarm_btn_long_press();
                break;
        }
    }

    if (_idle_ms && (millis() > _idle_ms)) {
        _mode = SHOW_TIME;  
    }
    
    return _mode;
}

bool option1() {
    return digitalRead(OPTION1) == LOW;
}

bool option2() {
    return digitalRead(OPTION2) == LOW;
}

bool option3() {
    return digitalRead(OPTION3) == LOW;
}

bool option4() {
    return digitalRead(OPTION4) == LOW;
}

unsigned char get_modified_value() {
    return _modified_value;
}
