/*
  DCF77.c - DCF77 library 
  Copyright (c) Thijs Elenbaas 2012

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  
  11 Apr 2012 - initial release 
  23 Apr 2012 - added UTC support
  2  Jul 2012 - minor bugfix and additional noise rejection
 */

#include "dcf77.h"       //https://github.com/thijse/Arduino-Libraries/downloads
#include "control.h"

#define _DCF77_VERSION 1_0_0 // software version of this library

#define DCF_PIN 2                // Connection pin to DCF 77 device
#define DCF_ACTIVE HIGH
#define DCF_LED_OK  30            
#define DCF_LED_ERR 31

#define MIN_TIME 1334102400     // Date: 11-4-2012
#define MAX_TIME 4102444800     // Date:  1-1-2100

#define DCFRejectionTime 700    // Pulse-to-Pulse rejection time. 
#define DCFMinPulseWidth 20  // Minimal pulse width
#define DCFMaxPulseWidth 260  // Maximal pulse width
#define DCFSplitTime 150        // Specifications distinguishes pulse width 100 ms and 200 ms. In practice we see 130 ms and 230
#define DCFSyncTime 1500        // Specifications defines 2000 ms pulse for end of sequence

DCF77 dcf;

void interruptHandler() {
    dcf.int0handler();
}

void dcf_init() {
    attachInterrupt(digitalPinToInterrupt(DCF_PIN), interruptHandler, CHANGE);
}

void dcf_stop() {
    detachInterrupt(digitalPinToInterrupt(DCF_PIN));
}

/**
 * Constructor
 */
DCF77::DCF77() {
    pinMode(DCF_PIN, INPUT_PULLUP);
    pinMode(DCF_LED_OK, OUTPUT);
    pinMode(DCF_LED_ERR, OUTPUT);

    // Parameters shared between interupt loop and main loop
    runningBuffer = 0;
    rxStatus = offline;
    telegramTime = 0;

    // DCF Buffers and indicators
    bitmask = 1;

    // Pulse flanks
    leadingEdgeTime = 0;
    trailingEdgeTime = 0;
}

/**
 * Initialize buffer for next time update
 */
inline void DCF77::InitBuffer(void) {
    runningBuffer = 0ULL;
    bitmask = 1ULL;
}

/**
 * Interrupt handler that processes up-down flanks into pulses and stores these in the buffer
 */
void DCF77::int0handler() {
    unsigned long edgeTime = millis();

    if (digitalRead(DCF_PIN) == DCF_ACTIVE) {
        unsigned long gap = edgeTime - leadingEdgeTime;
        if (gap > DCFRejectionTime) {
            unsigned long pulseWidth = trailingEdgeTime - leadingEdgeTime;
            if (pulseWidth < DCFSplitTime) {
                appendSignal(0);
            } else {
                appendSignal(1);
            }
            leadingEdgeTime = edgeTime;
        }
        if (gap > DCFSyncTime) {
            finalizeBuffer();
        }

        if (option1()) {
            switch (rxStatus) {
                case offline:
                    digitalWrite(DCF_LED_ERR, 1);
                    break;
                case online:
                    digitalWrite(DCF_LED_ERR, 1);
                    digitalWrite(DCF_LED_OK, 1);
                    break;
                case confirmed:
                    digitalWrite(DCF_LED_OK, 1);
                    break;
            }
        }
    } else {
        digitalWrite(DCF_LED_OK, 0);
        digitalWrite(DCF_LED_ERR, 0);
        trailingEdgeTime = edgeTime;
    }
}

/**
 * Add new bit to buffer
 */
inline void DCF77::appendSignal(unsigned char signal) {
    if (signal)
        runningBuffer |= bitmask;
    bitmask <<= 1;
}

/**
 * Finalize filled buffer
 */
inline void DCF77::finalizeBuffer(void) {
    if (bitmask == 0x800000000000000ULL && checkParity(runningBuffer)) {
        receivedTelegram.raw = runningBuffer;
        unsigned long thisMinute = 60 * receivedTelegram.value.Hour + receivedTelegram.value.Min;
        if (lastMinute + 1 == thisMinute && lastDay == receivedTelegram.value.Day && lastMonth == receivedTelegram.value.Month && lastYear == receivedTelegram.value.Year) {
            rxStatus = confirmed;
            telegramTime = millis();
        } else {
            rxStatus = online;
        }
        lastMinute = thisMinute;
        lastDay = receivedTelegram.value.Day;
        lastMonth = receivedTelegram.value.Month;
        lastYear = receivedTelegram.value.Year;
    } else {
        rxStatus = offline;
    }
    InitBuffer();
}

bool DCF77::checkParity(unsigned long long buffer) {
    bool evenParity = true;
    for (int pos = 21; pos < 29; pos++) {
        if ((buffer >> pos) & 1) {
            evenParity = !evenParity;
        }
    }

    if (!evenParity)
        return false;
    
    for (int pos = 29; pos < 36; pos++) {
        if ((buffer >> pos) & 1) {
            evenParity = !evenParity;
        }
    }

    if (!evenParity)
        return false;

    for (int pos = 36; pos < 59; pos++) {
        if ((buffer >> pos) & 1) {
            evenParity = !evenParity;
        }
    }

    return evenParity;
}
