/*
* DS1307.cpp - Library to set & get time from RTC DS1307
* Created by Manjunath CV. August 15, 2017, 07:00 PM
* Released into the public domain.
*/

#include <Arduino.h>
#include <time.h>
#include <Wire.h>
#include "ds1307.h"

bool DS1307::begin() {
  	Wire.begin(); // join i2c bus
  	Wire.beginTransmission(DS1307_ADDR);
  	return (Wire.endTransmission() == 0 ? true : false);
}

bool DS1307::isRunning() {
  	uint8_t data;
  	bool flag;
  	Wire.beginTransmission(DS1307_ADDR);
  	Wire.write(0x00);
  	Wire.endTransmission();
  	Wire.requestFrom(DS1307_ADDR, 1);
  	data = Wire.read();
  	flag = bitRead(data, 7);
  	return (!flag);
}

void DS1307::startClock() {
  	uint8_t data;
  	Wire.beginTransmission(DS1307_ADDR);
  	Wire.write(0x00);
  	Wire.endTransmission();
  	Wire.requestFrom(DS1307_ADDR, 1);
  	data = Wire.read();
  	bitClear(data, 7);
  	Wire.beginTransmission(DS1307_ADDR);
  	Wire.write(0x00);  // Seconds Register
  	Wire.write(data);
  	Wire.endTransmission();
}

void DS1307::stopClock() {
  	uint8_t data;
  	Wire.beginTransmission(DS1307_ADDR);
  	Wire.write(0x00);
  	Wire.endTransmission();
  	Wire.requestFrom(DS1307_ADDR, 1);
  	data = Wire.read();
  	bitSet(data, 7);
  	Wire.beginTransmission(DS1307_ADDR);
  	Wire.write(0x00);  // Seconds Register
  	Wire.write(data);
  	Wire.endTransmission();
}

/*-----------------------------------------------------------
get & set HourMode
-----------------------------------------------------------*/

void DS1307::setHourMode(uint8_t h_mode) {
  	uint8_t data;
  	if(h_mode == CLOCK_H12 || h_mode == CLOCK_H24) {
    		Wire.beginTransmission(DS1307_ADDR);
    		Wire.write(0x02);  // Hour Register
    		Wire.endTransmission();
    
    		Wire.requestFrom(DS1307_ADDR, 1);
    		data = Wire.read();
    		
    		bitWrite(data, 6, h_mode);
    		
    		Wire.beginTransmission(DS1307_ADDR);
    		Wire.write(0x02);  // Hour Register
    		Wire.write(data);
    		Wire.endTransmission();
  	}
}

uint8_t DS1307::getHourMode() {
  	bool h_mode;
  	uint8_t data;
  
  	Wire.beginTransmission(DS1307_ADDR);
  	Wire.write(0x02);
  	Wire.endTransmission();
  
  	Wire.requestFrom(DS1307_ADDR, 1);
  	data = Wire.read();
  
  	h_mode = bitRead(data, 6);
  
  	return (h_mode);
}

/*-----------------------------------------------------------
-----------------------------------------------------------*/

void DS1307::setMeridiem(uint8_t meridiem) {
  	uint8_t data;
  	if(meridiem == HOUR_AM || meridiem == HOUR_PM) {
    		if (getHourMode() == CLOCK_H12) {
      			Wire.beginTransmission(DS1307_ADDR);
      			Wire.write(0x02);  // Hour Register
      			Wire.endTransmission();
      			Wire.requestFrom(DS1307_ADDR, 1);
      			data = Wire.read();
      			bitWrite(data, 5, meridiem);
      			Wire.beginTransmission(DS1307_ADDR);
      			Wire.write(0x02);  // Hour Register
      			Wire.write(data);
      			Wire.endTransmission();
    		}
  	}
}

uint8_t DS1307::getMeridiem() {
  	bool flag;
  	uint8_t data;
  	if (getHourMode() == CLOCK_H12)   	{
    		Wire.beginTransmission(DS1307_ADDR);
    		Wire.write(0x02);
    		Wire.endTransmission();
    
    		Wire.requestFrom(DS1307_ADDR, 1);
    		data = Wire.read();
    
    		flag = bitRead(data, 5);
    		return (flag);
  	} else
    		return (HOUR_24);
}

/*-----------------------------------------------------------
get & set Second
-----------------------------------------------------------*/
uint8_t DS1307::getSeconds()
{
	volatile uint8_t seconds;
	Wire.beginTransmission(DS1307_ADDR);
	Wire.write(0x00);
	Wire.endTransmission();
	
	Wire.requestFrom(DS1307_ADDR, 1);
	seconds = Wire.read();
	bitClear(seconds, 7); // Clearing CH Bit if Set.
	return seconds;
}

void DS1307::setSeconds(uint8_t seconds)
{
  	uint8_t data, ch_bit;
		Wire.beginTransmission(DS1307_ADDR);
		Wire.write(0x00);  // Second Register
		Wire.endTransmission();
		Wire.requestFrom(DS1307_ADDR, 1);
		data = Wire.read();

		ch_bit = bitRead(data, 7);
		bitWrite(seconds,7,ch_bit);

		Wire.beginTransmission(DS1307_ADDR);
		Wire.write(0x00);  // Second Register
		Wire.write(seconds);
		Wire.endTransmission();
}

uint8_t DS1307::getMinutes() {
    return readRegister(0x01);
}

void DS1307::setMinutes(uint8_t minutes) {
    writeRegister(0x01, minutes);
}

uint8_t DS1307::getHours() {
  	uint8_t hour;
  	bool h_mode;
  	h_mode = getHourMode();
  	hour = readRegister(0x02);
  	if (h_mode == CLOCK_H24)
  	{
    		return hour;
  	}
  	else if (h_mode == CLOCK_H12)
  	{
    		bitClear(hour, 5);
    		bitClear(hour, 6);
    		return hour;
  	}
}

void  DS1307::setHours(uint8_t hours) {
	  bool h_mode = getHourMode();
		if (h_mode == CLOCK_H12)
		{
  			if (hours > 0x12)
  			{
    				hours -= 0x12;
    				bitSet(hours, 6);
    				bitSet(hours, 5);
  			}
  			else
  			{
    				bitSet(hours, 6);
    				bitClear(hours, 5);
  			}
		}

   writeRegister(0x02, hours);
}

/*-----------------------------------------------------------
getWeek
-----------------------------------------------------------*/
uint8_t DS1307::getWeek() {
    return readRegister(0x03);
}

void DS1307::setWeek(uint8_t week) {
  	if (week >= 1 && week <= 7) {
        writeRegister(0x03, week);
  	}
}

void DS1307::updateWeek() {
  	uint16_t y;
  	uint8_t m, d, weekday;
  	
  	y=getYear();
  	m=getMonth();
  	d=getDay();
  	
  	weekday  = (d += m < 3 ? y-- : y - 2, 23*m/9 + d + 4 + y/4- y/100 + y/400)%7;
  	
  	if (weekday >= 1 && weekday <= 7)
  	{
        writeRegister(0x03, weekday);
  	}
}

uint8_t DS1307::getDay() {
    return readRegister(0x04);
}

void DS1307::setDay(uint8_t day) {
    writeRegister(0x04, day);
}

uint8_t DS1307::getMonth() {
    return readRegister(0x05);
}

void DS1307::setMonth(uint8_t month) {
    writeRegister(0x05, month);
}

uint8_t DS1307::getYear() {
    return readRegister(0x06);
}

void DS1307::setYear(uint8_t year) {	
    writeRegister(0x06, year);
}

void DS1307::setAlarmHour(uint8_t hours) {
    writeRegister(0x08, hours);
}

uint8_t DS1307::getAlarmHour() {
    return readRegister(0x08);
}

void DS1307::setAlarmMinute(uint8_t minutes) {
    writeRegister(0x09, minutes);
}

uint8_t DS1307::getAlarmMinute() {
    return readRegister(0x09);
}

void DS1307::setAlarmSecond(uint8_t seconds) {
    writeRegister(0x0A, seconds);
}

uint8_t DS1307::getAlarmSecond() {
    return readRegister(0x0A);
}

void DS1307::setAlarmFlags(uint8_t flags) {
    writeRegister(0x0B, flags);
}

uint8_t DS1307::getAlarmFlags() {
    return readRegister(0x0B);
}

/* SQW/OUT pin functions */

void DS1307::setOutPin(uint8_t mode)
{
	Wire.beginTransmission(DS1307_ADDR);
	Wire.write(0x07);
	switch (mode) {
	case HIGH:
		Wire.write(B10000000);
		break;
	case LOW:
		Wire.write(B00000000);
		break;
	case SQW001Hz:
		Wire.write(B00010000);
		break;
	case SQW04kHz:
		Wire.write(B00010001);
		break;
	case SQW08kHz:
		Wire.write(B00010010);
		break;
	case SQW32kHz:
		Wire.write(B00010011);
		break;
	}
	Wire.endTransmission();
}

bool DS1307::isOutPinEnabled()
{
	uint8_t data;
	Wire.beginTransmission(DS1307_ADDR);
	Wire.write(0x07);
	Wire.endTransmission();
	Wire.requestFrom(DS1307_ADDR, 1);
	data = Wire.read();
	data = bitRead(data, 7);
	return (data);
}

bool DS1307::isSqweEnabled()
{
	uint8_t data;
	Wire.beginTransmission(DS1307_ADDR);
	Wire.write(0x07);
	Wire.endTransmission();
	Wire.requestFrom(DS1307_ADDR, 1);
	data = Wire.read();
	data = bitRead(data, 4);
	return (data);
}

uint8_t inline DS1307::readRegister(uint8_t address) {
    Wire.beginTransmission(DS1307_ADDR);
    Wire.write(address);  // Year Register
    Wire.endTransmission();
    Wire.requestFrom(DS1307_ADDR, 1);
    return Wire.read();
}

void inline DS1307::writeRegister(uint8_t address, uint8_t value) {
    Wire.beginTransmission(DS1307_ADDR); 
    Wire.write(address);  // Year Register to write year
    Wire.write(value);
    Wire.endTransmission();
}

/* NVRAM Functions */

bool NVRAM::begin()
{
	Wire.begin(); // join i2c bus
	Wire.beginTransmission(0x68);
	return (Wire.endTransmission() == 0 ? true : false);
}


uint8_t NVRAM::read(uint8_t address)
{
	uint8_t data = 0x00;
	address = (address % length) + 0x08;
	Wire.beginTransmission(DS1307_ADDR);
	Wire.write(address);
	Wire.endTransmission();
	Wire.requestFrom(DS1307_ADDR, 1);
	data = Wire.read();
	return (data);
}
void NVRAM::write(uint8_t address, uint8_t data)
{
	address = (address % length) + 0x08;
	Wire.beginTransmission(DS1307_ADDR);
	Wire.write(address);
	Wire.write(data);
	Wire.endTransmission();
}

void NVRAM::read(uint8_t address, uint8_t* buf, uint8_t size)
{
	int addrByte = 0x08 + address;
	Wire.beginTransmission(DS1307_ADDR);
	Wire.write(addrByte);
	Wire.endTransmission();
	Wire.requestFrom(DS1307_ADDR, 9);
	for (uint8_t pos = 0; pos < size; ++pos)
	{
		buf[pos] = Wire.read();
	}
}

void NVRAM::write(uint8_t address, uint8_t* buf, uint8_t size)
{
	int addrByte = 0x08 + address;
	Wire.beginTransmission(DS1307_ADDR);
	Wire.write(addrByte);
	for (uint8_t pos = 0; pos < size; ++pos)
	{
		Wire.write(buf[pos]);
	}
	Wire.endTransmission();
}

/* Helpers */

uint8_t DS1307::bcd2bin(uint8_t val)
{
	return val - 6 * (val >> 4);
}
uint8_t DS1307::bin2bcd(uint8_t val)
{
	return val + 6 * (val / 10);
}
