#include <Arduino.h>
#include "speaker.h"

// Timer 4
#define CUCKOO 6 // board PIN D6 - OC4A
#define BUZZER 7 // board PIN D7 - OC4B

// pattern: frequency [Hz], duration [16us]
typedef struct {
    unsigned short pitch;
    unsigned short duration;
} Note;

static const Note cucoo1[] = {719, 7125, 0, 9250, 576, 11812, 0, 53250};
static const Note cucoo2[] = {625, 7125, 0, 9250, 501, 11812, 0, 53250};
static const Note buzzer[] = {100, 62500, 0, 62500};

typedef struct {
    const Note* pattern;
    unsigned char pattern_length;
    unsigned char repeat_count;
    unsigned char output;
} Sequence;

static Sequence sequences[3];
static Sequence* sequence;

static unsigned char _note;

static void inline repeat_motive();
static void inline play_note();

void speaker_init() {
    cli();//stop interrupts
    TCCR4A = 0b00000000; // 00 channel A off, 00 channel B off, 00 channel C off, 00 mode4
    TCCR4B = 0;
    TCNT4  = 0;//initialize counter value to 0
    OCR4A = 0xFFFF; // 1 ~ 16 us
    // enable timer compare interrupt
    TIMSK4 |= (1 << OCIE4A);
    sei();//allow interrupts
}

static inline void stop_timer() {
    TCCR4B = 0;
}

static inline void start_timer(unsigned short delay) {
    TCNT4  = 0;//initialize counter value to 0
    OCR4A = delay;
    TCCR4B = 0b00001100;// 000 unused, 01 mode 4, 100 internal clock prescaler 1/256
}

static void inline next_motive() {
    sequence++;
    if (sequence->pattern) {
        repeat_motive();
    } else {
        stop_timer();
        pinMode(sequence->output,INPUT); 
    }
}

static void inline repeat_motive() {
    pinMode(sequence->output,OUTPUT); 
    if (sequence->repeat_count) {
        sequence->repeat_count--;
        _note = 0;
        play_note();
    } else {
        noTone(sequence->output);
        next_motive();
    }
}

static void inline play_note() {
    if (_note < sequence->pattern_length) {
        unsigned short pitch = sequence->pattern[_note].pitch;
        if (pitch) {
            tone(sequence->output, pitch);
        } else {
            noTone(sequence->output);
        }
        start_timer(sequence->pattern[_note].duration);
        _note++;
    } else {
        repeat_motive();
    }
}

static void inline playsequences() {
    sequence = sequences;
    repeat_motive();
}

void cuckoo(unsigned char count1, unsigned char count2) {
    sequences[0].pattern = cucoo1;
    sequences[0].pattern_length = 4;
    sequences[0].repeat_count = count1;
    sequences[0].output = CUCKOO;

    sequences[1].pattern = cucoo2;
    sequences[1].pattern_length = 4;
    sequences[1].repeat_count = count2;
    sequences[1].output = CUCKOO;
    
    sequences[2].pattern = NULL;

    playsequences();
}

void cuckoo1(unsigned char count) {
    sequences[0].pattern = cucoo1;
    sequences[0].pattern_length = 4;
    sequences[0].repeat_count = count;
    sequences[0].output = CUCKOO;

    sequences[1].pattern = NULL;

    playsequences();
}

void cuckoo2(unsigned char count) {
    sequences[0].pattern = cucoo2;
    sequences[0].pattern_length = 4;
    sequences[0].repeat_count = count;
    sequences[0].output = CUCKOO;

    sequences[1].pattern = NULL;

    playsequences();
}

void alarm(unsigned char count) {
    sequences[0].pattern = buzzer;
    sequences[0].pattern_length = 2;
    sequences[0].repeat_count = count;
    sequences[0].output = BUZZER;

    sequences[1].pattern = NULL;

    playsequences();
}

void speaker_mute() {
    sequence->repeat_count = 0;
    noTone(sequence->output);
}

ISR(TIMER4_COMPA_vect) {
    play_note();
}
